# -*- coding: utf-8 -*-
import datetime
import numpy as np
import pandas as pd
from beautifulexcel import ExcelWriter



def test_util_get_cell_coordinates():
    example_df = pd.DataFrame(
        {
            "client": ["A", "B", "C", "D", "E", "F", "G", "H"],
            "industry": [
                "ASEET MANAGEMENT",
                "ASEET MANAGEMENT",
                "BANK",
                "INSURANCE",
                "VERY VERY VERY VERY VERY LONG INSUSTRY NAME",
                "BANK",
                "COMMODITY BROKER",
                "INSURANCE",
                ],
            "employees": [25_000, 17_000_000, 14, 9_000, 12_000_000, 9_000, np.nan, 4_000_000],
            "inception": [
                datetime.datetime(2022, 1, 1),
                datetime.datetime(2000, 10, 30),
                datetime.datetime(2010, 5, 6),
                np.nan,
                datetime.datetime(1997, 1, 1),
                datetime.datetime(1962, 1, 1),
                datetime.datetime(2003, 11, 10),
                datetime.datetime(2022, 1, 1),
                ],
            "last_contact": [
                datetime.datetime(2022, 1, 1, 10, 2, 5),
                datetime.datetime(2000, 10, 30),
                datetime.datetime(2010, 5, 6),
                np.nan,
                datetime.datetime(1997, 1, 1),
                datetime.datetime(1962, 1, 1),
                datetime.datetime(2003, 11, 10, 10, 2, 5),
                datetime.datetime(2022, 1, 1),
                ],
            "RoE": [0.05, -0.05, 0.15, 1.05, -0.02, np.nan, 0.08, 0.05],
            "revenue": [
                50_000_000.4387,
                np.nan,
                63_000_000.4387,
                25_000.4387,
                25_000_000.4387,
                -50_000_000.4387,
                76_000_000.4387,
                25_000.4387,
                ],
            }
        )
    example_df.set_index("client", inplace=True)


    with ExcelWriter("testing.xlsx", mode="r", theme="elegant_blue") as writer:
        ws1 = writer.to_excel(
            example_df,
            sheet_name="Test Sheet 1",
            startrow=0,
            startcol=0,
            index=True,
            header=True,
        )

        assert ws1.startrow == 0
        assert ws1.startcol == 0
        assert ws1.index_depth == 1
        assert ws1.header_depth == 1
        assert ws1.table_width == example_df.shape[1]
        assert ws1.table_height == example_df.shape[0]
        assert ws1.shape == ((1, 1), (10, 8))
        assert ws1.shape_header == ((1, 1), (2, 8))
        assert ws1.shape_index == ((2, 1), (10, 2))
        assert ws1.shape_body == ((2, 2), (10, 8))

        assert (2, 1) == ws1.util_get_cell_coordinates(ref='A2')
        assert (None, 1) == ws1.util_get_cell_coordinates(ref='A')
        assert (2, None) == ws1.util_get_cell_coordinates(ref='2')
        assert (None, 3) == ws1.util_get_cell_coordinates(ref='employees')

        assert ((2, 1), (4, 3)) == ws1.util_get_range_coordinates('A2:C4')
        assert ((None, 1), (None, 3)) == ws1.util_get_range_coordinates('A:C')
        assert ((1, None), (3, None)) == ws1.util_get_range_coordinates('1:3')
        assert ((None, 3), (None, 3)) == ws1.util_get_range_coordinates(ref='employees')
        assert ((None, 3), (None, 5)) == ws1.util_get_range_coordinates(ref='employees:last_contact')

        assert 'A2:C4' == ws1.util_range_ref_from_coordinates(((2, 1), (4, 3)))

        ws2 = writer.to_excel(
            example_df,
            sheet_name="Test Sheet 2",
            startrow=2,
            startcol=3,
            index=False,
            header=False,
           )

        assert (2, 1) == ws2.util_get_cell_coordinates(ref='A2')
        assert (None, 1) == ws2.util_get_cell_coordinates(ref='A')
        assert (2, None) == ws2.util_get_cell_coordinates(ref='2')
        assert (None, 5) == ws2.util_get_cell_coordinates(ref='employees')







# -*- coding: utf-8 -*-
import re
import pandas as pd


def dict_extend_with_dict(dict_obj, key, value_dict):
    if key not in dict_obj:
        dict_obj[key] = {**value_dict}
    else:
        dict_obj[key] = {**dict_obj[key], **value_dict}
    return dict_obj


def flatten_dict(dict_obj: dict, sep="__"):
    """Flatten a multi-level dictionary e.g. {'a1': {'b': 1}, 'a2': 3} -> {'a1__b': 1, 'a2': 3}"""
    df = pd.json_normalize(dict_obj, sep=sep)
    d_flat = df.to_dict(orient="records")[0]
    return {k[2:] if len(k) > 2 and k[:2] == '__' else k: v for k, v in d_flat.items()}


def deepen_dict(dict_obj, sep="__"):
    """The opposite of flatten_dict e.g. {'a1__b': 1, 'a2': 3} -> {'a1': {'b': 1}, 'a2': 3}"""
    parsed_row = {}
    for col_label, v in dict_obj.items():
        keys = col_label.split(sep)

        current = parsed_row
        for i, k in enumerate(keys):
            if i == len(keys) - 1:
                current[k] = v
            else:
                if k not in current.keys():
                    current[k] = {}
                current = current[k]
    return parsed_row


def excel_column_name(n):
    """Number to Excel-style column name, e.g., 1 = A, 26 = Z, 27 = AA, 703 = AAA."""
    name = ""
    while n > 0:
        n, r = divmod(n - 1, 26)
        name = chr(r + ord("A")) + name
    return name


def excel_column_number(name):
    """Excel-style column name to number, e.g., A = 1, Z = 26, AA = 27, AAA = 703."""
    n = 0
    for c in name:
        n = n * 26 + 1 + ord(c) - ord("A")
    return n


def is_valid_excel_cell(cell):
    """Check if valid single excel cell ref like 'B2', 'AB100' etc."""
    m = re.match(r"^([A-Z]{1,2}|[A-W][A-Z]{2}|X[A-E][A-Z]|XF[A-D])([1-9]\d{0,6})$", cell)
    valid = bool(m) and int(m.group(2)) < 1_048_577
    return (valid, (m.group(1) if valid else None), (int(m.group(2)) if valid else None))


def excel_cell_ref_coordinates(ref, header=None, offset_horizontal=0, offset_vertical=0):
    """Transform references like 'A', 'COLUMN_A', '2', 'C2' to (row_num, col_num)"""
    # if already tuple - nothing to do
    if type(ref) is tuple:
        return ref

    # check if (relative) row number
    if ref.isdigit():
        return (int(ref) - 1 + offset_vertical, None)

    # check if column name
    if header is not None and ref in header:
        return (None, header.get_loc(ref) + offset_horizontal)

    # check if (reasonable) column letter
    if len(ref) < 4 and ref.isalpha() and excel_column_number(ref) < 1_048_577:
        return (None, excel_column_number(ref) - 1)

    # check if valid excel cell ref
    valid_excel_cell, col_letter, row_num = is_valid_excel_cell(ref)
    if valid_excel_cell:
        return (row_num - 1, excel_column_number(col_letter) - 1)


def excel_range_ref_coordinates(ref, header, offset_horizontal=0, offset_vertical=0):
    """Transform references like 'A:C', 'COLUMN_A:COLUMN_C', '2:3', 'C2:D4' to ((row_num, col_num), (row_num, col_num))"""
    if type(ref) is tuple:
        ref_start, ref_end = ref

    elif type(ref) is str:
        # ref is range
        if ":" in ref:
            ref_start, ref_end = ref.split(":")
        # ref is no range
        else:
            ref_start = ref_end = ref

    ref_start = excel_cell_ref_coordinates(
        ref=ref_start, header=header, offset_horizontal=offset_horizontal, offset_vertical=offset_vertical
    )
    ref_end = excel_cell_ref_coordinates(
        ref=ref_end, header=header, offset_horizontal=offset_horizontal, offset_vertical=offset_vertical
    )

    if ref_start is None or ref_end is None:
        # cell range is invalid
        return (ref_start, ref_end)
    else:
        # ensure valid ascending cell ref - e.g. C1:A5 -> A1:C5
        row_start, col_start = ref_start
        row_end, col_end = ref_end

        if row_start is not None and row_end is not None and row_end < row_start:
            row_start, row_end = row_end, row_start
        if col_start is not None and col_end is not None and col_end < col_start:
            col_start, col_end = col_end, col_start

        return ((row_start, col_start), (row_end, col_end))


def cell_within_cell_range(cell_row_num, cell_col_num, cell_range, ignore_entire_rows_or_cols=False):
    """Check if a cell e.g. 'A1' (0, 0) is within a cell range e.g. ((0, 0), (1, 1))"""
    ((range_start_row, range_start_col), (range_end_row, range_end_col)) = cell_range

    # if ignore_entire_rows_or_cols is True
    if ignore_entire_rows_or_cols and any(
        [range_start_row is None, range_start_col is None, range_end_row is None, range_end_col is None]
    ):
        return False

    # outside on the top
    if range_start_row is not None and range_start_row > cell_row_num:
        return False

    # outside on the bottom
    if range_end_row is not None and range_end_row < cell_row_num:
        return False

    # outside on the left
    if range_start_col is not None and range_start_col > cell_col_num:
        return False

    # outside on the right
    if range_end_col is not None and range_end_col < cell_col_num:
        return False

    # inside
    return True


def get_custom_styles(cell_row_num, cell_col_num, styles_custom, ignore_entire_rows_or_cols=False):
    """Check if custom style needs to be applied to current cell and return custom style"""
    cell_style = {}
    for cell_range, cust_style in styles_custom.items():
        # if current cell within cell range of cutsom style add style
        if cell_within_cell_range(
            cell_row_num=cell_row_num,
            cell_col_num=cell_col_num,
            cell_range=cell_range,
            ignore_entire_rows_or_cols=ignore_entire_rows_or_cols,
        ):
            cell_style = {**cell_style, **cust_style}

    return cell_style



# -*- coding: utf-8 -*-
import pandas as pd
from beautifulexcel.utils import excel_cell_ref_coordinates, excel_range_ref_coordinates, cell_within_cell_range


def test_cell_single_level_df():
    df = pd.DataFrame({"calories": [420, 380, 390, 210], "duration": [50, 40, 45, 90], "test": [4, 1, 4, 7]})

    assert (None, 1) == excel_cell_ref_coordinates("duration", df.columns)
    assert (None, 1) == excel_cell_ref_coordinates("B", df.columns)
    assert (0, None) == excel_cell_ref_coordinates("1", df.columns)
    assert (1, 1) == excel_cell_ref_coordinates("B2", df.columns)
    assert (2, 2) == excel_cell_ref_coordinates((2, 2), df.columns)


def test_range_single_level_df():
    df = pd.DataFrame({"calories": [420, 380, 390, 210], "duration": [50, 40, 45, 90], "test": [4, 1, 4, 7]})

    assert ((None, 1), (None, 2)) == excel_range_ref_coordinates(
        "duration:test", df.columns, offset_horizontal=0, offset_vertical=0
    )
    assert ((None, 1), (None, 2)) == excel_range_ref_coordinates(
        "duration:test", df.columns, offset_horizontal=0, offset_vertical=1
    )
    assert ((None, 2), (None, 3)) == excel_range_ref_coordinates(
        "duration:test", df.columns, offset_horizontal=1, offset_vertical=0
    )
    assert ((0, None), (2, None)) == excel_range_ref_coordinates(
        "1:3", df.columns, offset_horizontal=0, offset_vertical=0
    )
    assert ((1, None), (3, None)) == excel_range_ref_coordinates(
        "1:3", df.columns, offset_horizontal=0, offset_vertical=1
    )
    assert ((0, None), (2, None)) == excel_range_ref_coordinates(
        "1:3", df.columns, offset_horizontal=1, offset_vertical=0
    )
    assert ((None, 1), (None, 2)) == excel_range_ref_coordinates(
        "B:C", df.columns, offset_horizontal=5, offset_vertical=5
    )
    assert ((0, 0), (1, 2)) == excel_range_ref_coordinates("A1:C2", df.columns, offset_horizontal=5, offset_vertical=5)
    assert ((1, 1), (5, 6)) == excel_range_ref_coordinates(((5, 1), (1, 6)), df.columns)
    assert ((0, 0), (1, 2)) == excel_range_ref_coordinates("A2:C1", df.columns, offset_horizontal=5, offset_vertical=5)
    assert ((0, 0), (1, 2)) == excel_range_ref_coordinates("C1:A2", df.columns, offset_horizontal=5, offset_vertical=5)


def test_cell_within_cell_range():
    cell_range = excel_range_ref_coordinates("B2:D4", None)

    a1 = excel_cell_ref_coordinates("A1", None)
    b1 = excel_cell_ref_coordinates("B1", None)
    c1 = excel_cell_ref_coordinates("C1", None)
    d1 = excel_cell_ref_coordinates("D1", None)
    e1 = excel_cell_ref_coordinates("E1", None)

    assert ~cell_within_cell_range(a1[0], a1[1], cell_range)
    assert ~cell_within_cell_range(b1[0], a1[1], cell_range)
    assert ~cell_within_cell_range(c1[0], a1[1], cell_range)
    assert ~cell_within_cell_range(d1[0], a1[1], cell_range)
    assert ~cell_within_cell_range(e1[0], e1[1], cell_range)

    b2 = excel_cell_ref_coordinates("B2", None)
    c3 = excel_cell_ref_coordinates("C3", None)
    d4 = excel_cell_ref_coordinates("D4", None)
    c2 = excel_cell_ref_coordinates("C2", None)
    d2 = excel_cell_ref_coordinates("D2", None)
    d3 = excel_cell_ref_coordinates("D3", None)
    b4 = excel_cell_ref_coordinates("B4", None)

    assert cell_within_cell_range(b2[0], b2[1], cell_range)
    assert cell_within_cell_range(c3[0], c3[1], cell_range)
    assert cell_within_cell_range(d4[0], d4[1], cell_range)
    assert cell_within_cell_range(c2[0], c2[1], cell_range)
    assert cell_within_cell_range(d2[0], d2[1], cell_range)
    assert cell_within_cell_range(d3[0], d3[1], cell_range)
    assert cell_within_cell_range(b4[0], b4[1], cell_range)

    a3 = excel_cell_ref_coordinates("A3", None)
    a4 = excel_cell_ref_coordinates("A4", None)
    a5 = excel_cell_ref_coordinates("A5", None)
    c5 = excel_cell_ref_coordinates("C5", None)
    d5 = excel_cell_ref_coordinates("D5", None)
    e5 = excel_cell_ref_coordinates("E5", None)
    e4 = excel_cell_ref_coordinates("E4", None)
    e3 = excel_cell_ref_coordinates("E3", None)

    assert ~cell_within_cell_range(a3[0], a3[1], cell_range)
    assert ~cell_within_cell_range(a4[0], a4[1], cell_range)
    assert ~cell_within_cell_range(a5[0], a5[1], cell_range)
    assert ~cell_within_cell_range(c5[0], c5[1], cell_range)
    assert ~cell_within_cell_range(d5[0], d5[1], cell_range)
    assert ~cell_within_cell_range(e5[0], e5[1], cell_range)
    assert ~cell_within_cell_range(e4[0], e4[1], cell_range)
    assert ~cell_within_cell_range(e3[0], e3[1], cell_range)

    col_range = excel_range_ref_coordinates("B:C", None)
    row_range = excel_range_ref_coordinates("2:3", None)

    a1 = excel_cell_ref_coordinates("A1", None)
    b2 = excel_cell_ref_coordinates("B2", None)
    c3 = excel_cell_ref_coordinates("C3", None)
    d4 = excel_cell_ref_coordinates("D4", None)

    assert ~cell_within_cell_range(a1[0], a1[1], col_range)
    assert cell_within_cell_range(b2[0], b2[1], col_range)
    assert cell_within_cell_range(c3[0], c3[1], col_range)
    assert ~cell_within_cell_range(d4[0], d4[1], col_range)

    assert ~cell_within_cell_range(a1[0], a1[1], row_range)
    assert cell_within_cell_range(b2[0], b2[1], row_range)
    assert cell_within_cell_range(c3[0], c3[1], row_range)
    assert ~cell_within_cell_range(d4[0], d4[1], row_range)



# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages
import datetime

__version__ = "0.3.5"
if "$" in __version__:
    __version__ = datetime.datetime.now().strftime("%Y.%m.%d.%H.%M")
print("Version:", __version__)

setup(
    name="beautifulexcel",
    version=__version__,
    description="BeautifulExcel is a python package that makes it easy and quick to save pandas dataframes in beautifully formatted excel files. BeautifulExcel is the Openpyxl for Data Scientists with a deadline.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="https://github.com/vanalmsick",
    url="https://github.com/vanalmsick/beautifulexcel",
    project_urls={
        "Issues": "https://github.com/vanalmsick/beautifulexcel/issues",
        "Documentation": "https://vanalmsick.github.io/beautifulexcel/",
        "Source Code": "https://github.com/vanalmsick/beautifulexcel",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    #packages=find_packages(),
    install_requires=['typing', 'PyYAML', 'datetime', 'openpyxl', 'numpy', 'pandas'],
    package_data={"beautifulexcel": ["themes/*.yml", "VERSION"]},
    include_package_data=True,
)



# -*- coding: utf-8 -*-
import os
import warnings
from typing import Any, Union, List, Tuple #, Literal
import datetime
import numpy as np
import pandas as pd
import openpyxl
import yaml


from beautifulexcel.utils import (
    flatten_dict,
    deepen_dict,
    get_custom_styles,
    dict_extend_with_dict,
    is_valid_excel_cell,
)


class Sheet:
    """Base Excel Sheet Class which contains all the methods that can be applied to a sheet"""

    def __init__(self, excelwriter, sheet_name, use_theme_style, col_widths):
        self.excelwriter = excelwriter
        self.writer = self.excelwriter.writer
        self.sheet_name = sheet_name
        self.use_theme_style = use_theme_style
        self.col_widths = col_widths

        if use_theme_style:
            self.style_base = self._extend_style_args(self.excelwriter.theme.get("general", {}))
        else:
            self.style_base = {}

    def _extend_style_args(self, style_args):
        """Uses the theme presets and replaces them with the actual style kwargs"""
        self.preset_styles = preset_styles = self.excelwriter.theme.get("preset", {})

        # already extended dict style
        if type(style_args) is dict:
            return style_args
        # single preset name
        elif type(style_args) is str:
            style_i = preset_styles.get(style_args)
            if style_i is None:
                raise Exception(f'You are referencing the style preset "{style_i}" but it is not defined in the theme.')
            return preset_styles.get(style_args)
        # list of preset names or extended dict
        elif type(style_args) is list:
            _final_style = {}
            for i_style_args in style_args:
                # is preset name
                if type(i_style_args) is str:
                    if style_i is None:
                        style_i = preset_styles.get(i_style_args)
                        raise Exception(
                            f'You are referencing the style preset "{i_style_args}" but it is not defined in the theme.'
                        )
                    _final_style = {**_final_style, **style_i}
                # is already extended dict
                else:
                    _final_style = {**_final_style, **i_style_args}
            return _final_style

    def apply_cell_style(self, row_num, col_num, style):
        ws = self.ws
        style_deep = deepen_dict(style)
        cell = ws.cell(row=row_num, column=col_num)

        for style_type, kwargs in style_deep.items():
            if style_type.lower() == "font":
                cell.font = openpyxl.styles.Font(**kwargs)
            elif style_type.lower() == "numfmt" or style_type.lower() == "numberformat":
                cell.number_format = kwargs
            elif style_type.lower() == "align" or style_type.lower() == "alignment":
                cell.alignment = openpyxl.styles.Alignment(**kwargs)
            elif style_type.lower() == "fill" or style_type.lower() == "pfill" or style_type.lower() == "patternfill":
                if type(kwargs) is not dict:
                    kwargs = {"patternType": "solid", "fgColor": kwargs}
                cell.fill = openpyxl.styles.PatternFill(**kwargs)
            elif style_type.lower() == "gfill" or style_type.lower() == "gradientfill":
                cell.fill = openpyxl.styles.GradientFill(**kwargs)
            elif style_type.lower() == "border" or style_type.lower() == "borders":
                border_kwargs = {}
                for border_type, border_props in kwargs.items():
                    border_kwargs[border_type] = openpyxl.styles.Side(**border_props)
                cell.border = openpyxl.styles.Border(**border_kwargs)
            elif style_type.lower() == "protection":
                cell.font = openpyxl.styles.Protection(**kwargs)
            else:
                raise Exception(
                    f'Unknown style type "{style_type}". Available style types are: font, numberformat, align, fill, patternfill, gradientfill, borders, and protection.'
                )

        return cell

    def apply_range_style(self, ws, style):
        raise Exception("Not yet defined function .apply_range_style()")

    def change_col_widths(self, col_widths: dict):
        for col, width in col_widths.items():
            if type(col) is int:
                col = openpyxl.utils.get_column_letter(col + 1)
            self.ws.column_dimensions[col].width = width

    def add_data_validation(
            self,
            ref: Union[str, List[str]],
            type: str, #Literal["list", "whole", "decimal", "date", "time", "textLength", "formula"],
            props: Union[str, List[str], int, float, Tuple[int, float]] = None,
            operator: str = None, #Literal["between", "greaterThan", "greaterThanOrEqual", "equal", "notEqual", "lessThan", "lessThanOrEqual"] = None,
            **kwargs
        ):
        """
        Add data validation to worksheet

        Args:
            ref (str or list of str): Cell range reference e.g. "A1:C5" or ["A1:C5", "A21:Z55"]
            type (str): Data valiation type. Options: list, whole, decimal, date, time, textLength, formula
            operator (str): Validation mathematical operator: "between", "greaterThan", "greaterThanOrEqual", "equal", "notEqual", "lessThan", "lessThanOrEqual"
            props (any): Depending on vaildation type the properties: list = ["Option 1", "Option 2"], numeric = single number OR upper & lower boundary (1, 100), formla
            kwargs (dict): Any futher kwarg specified in [openpyxl.readthedocs.io/en/stable/api/openpyxl.worksheet.datavalidation.html](https://openpyxl.readthedocs.io/en/stable/api/openpyxl.worksheet.datavalidation.html#openpyxl.worksheet.datavalidation.DataValidation)

        Examples:
            - ```sheet1.add_data_validation(ref="A1:C5", type="whole")```
            - ```sheet1.add_data_validation(ref="A1:C5", type="decimal", operator="greaterThan", props=0)```
            - ```sheet1.add_data_validation(ref="A1:C5", type="textLength", props=10)```
            - ```sheet1.add_data_validation(ref="A1:C5", type="list", props=["Yes", "No"])```
            - ```sheet1.add_data_validation(ref="A1:C5", type="whole", operator="between", props=[0, 100])```
        """
        if isinstance(props, list) or isinstance(props, tuple):
            if type.lower() == 'list':
                formula1 = '"' + ','.join(props) + '"'
                formula2 = None
            else:
                formula1, formula2 = props
        else:
            formula1 = props
            formula2 = None
        dv = openpyxl.worksheet.datavalidation.DataValidation(type=type, formula1=formula1, formula2=formula2, operator=operator, **kwargs)
        self.ws.add_data_validation(dv)

        if isinstance(ref, str):
            ref = [ref]
        for i_ref in ref:
            if self.__class__.__name__ == 'DataframeSheet':
                i_ref = self.util_range_ref_from_coordinates(self.util_get_range_coordinates(i_ref, enrich_dimensions=True))
            if i_ref is not None:
                dv.add(i_ref)


    def merge_cells(self, ref: str):
        """
        Merge several cells

        Args:
            ref (str): Cell range reference e.g. 'A1:C5'

        Example:
            ```sheet1.merge_cells(ref='A1:C5')```
        """
        if isinstance(ref, str):
            ref = [ref]
        for i_ref in ref:
            if self.__class__.__name__ == 'DataframeSheet':
                i_ref = self.util_range_ref_from_coordinates(self.util_get_range_coordinates(i_ref, enrich_dimensions=True))
            if i_ref is not None:
                self.ws.merge_cells(i_ref)


    def __group_cols_rows(self, ref, axis, hidden=False, outline_level=1):
        """Group several rows or cols"""
        if isinstance(ref, str):
            ref = [ref]
            if isinstance(hidden, bool):
                hidden = [hidden] * len(ref)
            if isinstance(outline_level, int):
                outline_level = [outline_level] * len(ref)
        for i_ref, i_hidden, i_level in zip(ref, hidden, outline_level):
            if self.__class__.__name__ == 'DataframeSheet':
                i_ref = self.util_range_ref_from_coordinates(self.util_get_range_coordinates(i_ref))
            if i_ref is not None:
                i_ref_start, i_ref_end = i_ref.split(':')
                if axis == 'columns':
                    self.ws.column_dimensions.group(i_ref_start, i_ref_end, hidden=i_hidden, outline_level=i_level)
                elif axis == 'rows':
                    self.ws.row_dimensions.group(i_ref_start, i_ref_end, hidden=i_hidden, outline_level=i_level)


    def group_columns(self, ref: str):
        """
        Group several columns

        Args:
            ref (str): Excel column reference e.g. 'A:C'

        Example:
            ```sheet1.group_columns(ref='1:2')```
        """
        return self.__group_cols_rows(ref=ref, axis='columns')

    def group_rows(self, ref):
        """
        Group several rows

        Args:
            ref (str): Excel row reference e.g. '1:3'

        Example:
            ```sheet1.group_rows(ref='1:5')```
        """
        return self.__group_cols_rows(ref=ref, axis='rows')

    def util_get_cell_coordinates(self, ref):
        """
        Get cell coordinates

        Examples:
            >>> ws.util_get_cell_coordinates(ref='A2')
            (2, 1)
            >>> ws.util_get_cell_coordinates(ref='A')
            (None, 1)
            >>> ws.util_get_cell_coordinates(ref='1')
            (1, None)
            >>> ws.util_get_cell_coordinates(ref='-1')
            (-1, None)
        """
        ref = str(ref)

        # check if row number - i.e. all numbers
        if ref.isdigit() or (len(ref) >= 2 and ref[0] == '-' and ref[1:].isdigit()):
            return (int(ref), None)

        # check if reasonable column letter - i.e. all letters and resonable number
        if len(ref) < 4 and ref.isalpha() and openpyxl.utils.cell.column_index_from_string(ref) < 1_048_577:
            return (None, openpyxl.utils.cell.column_index_from_string(ref))

        # check if valid excel cell ref
        valid_excel_cell, col_letter, row_num = is_valid_excel_cell(ref)
        if valid_excel_cell:
            return openpyxl.utils.cell.coordinate_to_tuple(ref)

        # raise wainring if invalid
        if self.excelwriter.ref_warnings:
            warnings.warn(f'Ref "{ref}" could not be found. (You can ignore these warnings by adding ref_warnings=False to ExcelWriter())')

        # return None
        return None


    def util_get_range_coordinates(self, ref, enrich_dimensions=False):
        """
        Get cell range coordinates

        Examples:
            >>> ws.util_get_range_coordinates(ref='A2:C4')
            ((2, 1), (4, 3))
            >>> ws.util_get_range_coordinates(ref='A:C')
            ((None, 1), (None, 3))
            >>> ws.util_get_range_coordinates(ref='1:3')
            ((1, None), (3, None))
            >>> ws.util_get_range_coordinates(ref='4:-1')
            ((4, None), (-1, None))

        """
        if ':' in ref:
            ref_start, ref_end = ref.split(':')
        else:
            ref_start = ref_end = ref

        coordinates_start = self.util_get_cell_coordinates(ref_start)
        coordinates_end = self.util_get_cell_coordinates(ref_end)

        # if any invaild ref
        if coordinates_start is None or coordinates_end is None:
            return None

        row_start, col_start = coordinates_start
        row_end, col_end = coordinates_end
        if self.__class__.__name__ == 'DataframeSheet':
            ((sheet_min_row, sheet_min_column), (sheet_max_row, sheet_max_column)) = self.shape_body
        else:
            sheet_min_row = self.ws.min_row
            sheet_max_row = self.ws.max_row
            sheet_min_column = self.ws.min_column
            sheet_max_column = self.ws.max_column

        # enrich_dimensions if any ref is None/undefined
        if enrich_dimensions:
            if row_start is None:
                row_start = sheet_min_row
            if row_end is None:
                row_end = sheet_max_row - 1
            if col_start is None:
                col_start = sheet_min_column
            if col_end is None:
                col_end = sheet_max_column - 1

        # make sure lower bound at start always - so range() works
        if row_start is not None and row_end is not None and row_end < row_start:
            row_end, row_start = row_start, row_end
        if col_start is not None and col_end is not None and col_end < col_start:
            col_end, col_start = col_start, col_end

        return ((row_start, col_start), (row_end, col_end))

    def util_cell_ref_from_coordinates(self, ref):
        """Tranform (row_num, col_num) into excel ref 'A1' """
        if ref is None:
            return None

        col_letter = '' if ref[1] is None else openpyxl.utils.cell.get_column_letter(ref[1])
        row_number = '' if ref[0] is None else ref[0]
        return f'{col_letter}{row_number}'

    def util_range_ref_from_coordinates(self, ref):
        """Tranform ((start_row_num, start_col_num), (end_row_num, end_col_num)) into excel ref 'A1:C5'"""
        if ref is None:
            return None

        start_ref, end_ref = ref
        start_ref = self.util_cell_ref_from_coordinates(start_ref)
        end_ref = self.util_cell_ref_from_coordinates(end_ref)
        return f'{start_ref}:{end_ref}'


    def write_cell(self, ref: str, content):
        self.ws[ref] = content

    def write_cells(self, cell_dict: dict):
        for ref, content in cell_dict.items():
            self.write_cell(ref=ref, content=content)


class DataframeSheet(Sheet):
    """
    DataFrame Excel Sheet class containing all logic specific to dataframe exports

    Note: All methods/function of the class Sheet also work for this class DataframeSheet
    """

    def __init__(
        self,
        excelwriter,
        df,
        sheet_name,
        startrow=0,
        startcol=0,
        index=False,
        header=True,
        style={},
        use_theme_style=True,
        col_widths={},
        col_autofit=True,
        auto_number_formatting=True,
    ):
        super().__init__(excelwriter, sheet_name, use_theme_style, col_widths)
        self.startrow = startrow
        self.startcol = startcol
        self.df = df
        self.index = df.index
        self.header = df.columns
        self.has_index = index
        self.has_header = header
        self.index_depth = index_depth = self.index.nlevels if self.has_index else 0
        self.header_depth = header_depth = self.header.nlevels if self.has_header else 0
        self.table_width = table_width = len(self.header)
        self.table_height = table_height = len(self.index)
        self.col_autofit = col_autofit
        self.auto_number_formatting = auto_number_formatting
        self.shape = ((startrow + 1, startcol + 1), (startrow + header_depth + table_height + 1, startcol + index_depth + table_width + 1))
        self.shape_header = ((startrow + 1, startcol + 1), (startrow + header_depth + 1, startcol + index_depth + table_width + 1))
        self.shape_index = ((startrow + header_depth + 1, startcol + 1), (startrow + header_depth + table_height + 1, startcol + index_depth + 1))
        self.shape_body = ((startrow + header_depth + 1, startcol + index_depth + 1), (startrow + header_depth + table_height + 1, startcol + index_depth + table_width + 1))

        # export df to excel
        self.df.to_excel(
            self.writer, sheet_name=sheet_name, startrow=startrow, startcol=startcol, index=index, header=header
        )
        self.ws = self.writer.book[sheet_name]

        # generate final styling that will apply to the dataframe export
        if use_theme_style:
            # add table style from style template
            if "table" in self.excelwriter.theme:
                for level, level_styling in self.excelwriter.theme["table"].items():
                    dict_extend_with_dict(
                        dict_obj=self.style_base, key=level, value_dict=self._extend_style_args(level_styling)
                    )

        # add number formatting
        if auto_number_formatting:
            # get all date columns
            for col_name, col_series in df.select_dtypes(include=["datetime", "datetimetz"]).items():
                self.style_base[col_name] = self._extend_style_args("date_fmt_iso")

            # get all numeric columns
            for col_name, col_series in df.select_dtypes(include=["number"]).items():
                col_series_mod = col_series.replace(0, np.nan).abs()
                low = col_series_mod.quantile(0.2)
                high = col_series_mod.quantile(0.8)

                # check if percentages
                if high < 2:
                    self.style_base[col_name] = self._extend_style_args("num_fmt_pct")
                # check if small number
                elif high < 1_000 and "int" not in str(col_series.dtype):
                    self.style_base[col_name] = self._extend_style_args("num_fmt_decimal")
                # check if large number in millions
                elif low > 10_000_000:
                    self.style_base[col_name] = self._extend_style_args("num_fmt_mm")
                # else normal number format
                else:
                    self.style_base[col_name] = self._extend_style_args("num_fmt_general")

        # add styling defined in this function
        self.style_custom = {}
        for ref, ref_style in style.items():
            dict_extend_with_dict(dict_obj=self.style_custom, key=ref, value_dict=self._extend_style_args(ref_style))

        # actually apply the final themes
        self._apply_table_style()

        # apply column widths
        _col_widths = {}
        for col_ref, col_width in col_widths.items():
            col_coordinates = self.util_get_range_coordinates(col_ref)

            if (
                col_coordinates is not None
                and col_coordinates[0] is not None
                and col_coordinates[0][1] is not None
                and col_coordinates[1] is not None
                and col_coordinates[1][1] is not None
            ):
                for col_idx in range(col_coordinates[0][1], col_coordinates[1][1] + 1):
                    _col_widths[col_idx] = col_width

        # autofit columns
        if col_autofit:
            CHARACTER_FACTOR = 1.28

            # enumerate though indeices and columns
            for i, col in enumerate(
                ([df.index.get_level_values(i) for i in range(df.index.nlevels)] if index else [])
                + [i for _, i in df.items()],
                start=startcol,
            ):
                # check if no manual col width
                if i not in _col_widths:
                    try:
                        # is datetime column
                        if pd.api.types.is_datetime64_any_dtype(col):
                            _col_widths[i] = int(10 * CHARACTER_FACTOR)
                        # is string column
                        elif pd.api.types.is_string_dtype(col):
                            try:
                                length = np.quantile(col.str.len().values, 0.8)
                                _col_widths[i] = int(max(length, 4) * CHARACTER_FACTOR)
                            except:
                                # ignore error
                                pass
                        # is numeric column
                        elif pd.api.types.is_numeric_dtype(col):
                            length = len("{:,}".format(int(col.quantile(0.8))))
                            _col_widths[i] = int(max(length, 6) * CHARACTER_FACTOR)
                    except Exception as e:
                        warning.warn(f'Were not able to automatically determin best column width for "{col}". You can set it manually col_widths=' + '{' + f'"{col}": 40' + '}')

        self.change_col_widths(_col_widths)

    def _apply_table_style(self):
        """This internal function applies the table cell styling for the to_excel() function"""
        ref_warnings = self.excelwriter.ref_warnings

        style_base = self.style_base.copy()
        style_custom = self.style_custom.copy()

        style_special_base = {**style_base.pop("base", {}), **style_custom.pop("base", {})}

        style_special_head = {**style_special_base, **style_base.pop("head", {}), **style_custom.pop("head", {})}
        style_special_index = {**style_special_base, **style_base.pop("index", {}), **style_custom.pop("index", {})}
        style_special_body = {**style_special_base, **style_base.pop("body", {}), **style_custom.pop("body", {})}

        style_non_special = {}
        for iter_dict in [style_base, style_custom]:
            for ref, ref_style in iter_dict.items():
                style_coordinates = self.util_get_range_coordinates(ref)

                # check if invalid cell ref
                if (style_coordinates is None or style_coordinates[0] is None or style_coordinates[1] is None):
                    if ref_warnings:
                        warnings.warn(
                            f'Styling ref "{ref}" could not be found. (You can turn off these warnings by adding ref_warnings=False to ExcelWriter())'
                        )

                # all good - valid cell style range
                else:
                    dict_extend_with_dict(
                        dict_obj=style_non_special, key=style_coordinates, value_dict=self._extend_style_args(ref_style)
                    )

        # Apply table heading styling
        if self.has_header:
            ((start_row, start_col), (end_row, end_col)) = self.shape_header
            for row_num in range(start_row, end_row):
                for col_num in range(start_col, end_col):
                    cell_style = get_custom_styles(
                        cell_row_num=row_num,
                        cell_col_num=col_num,
                        styles_custom=style_non_special,
                        ignore_entire_rows_or_cols=True,
                    )
                    # print('Header', row_num, col_num)
                    self.apply_cell_style(row_num=row_num, col_num=col_num, style={**style_special_head, **cell_style})

        # Apply table index styling
        if self.has_index:
            ((start_row, start_col), (end_row, end_col)) = self.shape_index
            for row_num in range(start_row, end_row):
                for col_num in range(start_col, end_col):
                    cell_style = get_custom_styles(
                        cell_row_num=row_num,
                        cell_col_num=col_num,
                        styles_custom=style_non_special,
                        ignore_entire_rows_or_cols=True,
                    )
                    # print('Index', row_num, col_num)
                    self.apply_cell_style(row_num=row_num, col_num=col_num, style={**style_special_index, **cell_style})

        # Apply table body styling
        ((start_row, start_col), (end_row, end_col)) = self.shape_body
        for row_num in range(start_row, end_row):
            for col_num in range(start_col, end_col):
                cell_style = get_custom_styles(
                    cell_row_num=row_num,
                    cell_col_num=col_num,
                    styles_custom=style_non_special,
                    ignore_entire_rows_or_cols=False,
                )
                # print('Body', row_num, col_num)
                self.apply_cell_style(row_num=row_num, col_num=col_num, style={**style_special_body, **cell_style})


    def add_data_validation(self,
            ref: Union[str, List[str]],
            type: str, #Literal["list", "whole", "decimal", "date", "time", "textLength", "formula"],
            props: Union[str, List[str], int, float, Tuple[int, float]] = None,
            operator: str = None, #Literal["between", "greaterThan", "greaterThanOrEqual", "equal", "notEqual", "lessThan", "lessThanOrEqual"] = None,
            **kwargs
        ):
        """
        Add data validation to worksheet

        Args:
            ref (str or list of str): Cell range reference e.g. "A1:C5" or "employees" or ["A1:C5", "A21:Z55", "employees:RoE"]
            type (str): Data valiation type. Options: list, whole, decimal, date, time, textLength, formula
            operator (str): Validation mathematical operator: "between", "greaterThan", "greaterThanOrEqual", "equal", "notEqual", "lessThan", "lessThanOrEqual"
            props (any): Depending on vaildation type the properties: list = ["Option 1", "Option 2"], numeric = single number OR upper & lower boundary (1, 100), formla
            kwargs (dict): Any futher kwarg specified in [openpyxl.readthedocs.io/en/stable/api/openpyxl.worksheet.datavalidation.html](https://openpyxl.readthedocs.io/en/stable/api/openpyxl.worksheet.datavalidation.html#openpyxl.worksheet.datavalidation.DataValidation)

        Examples:
            - ```sheet1.add_data_validation(ref="employees", type="whole")```
            - ```sheet1.add_data_validation(ref="A1:C5", type="whole")```
            - ```sheet1.add_data_validation(ref="RoE", type="decimal", operator="greaterThan", props=0)```
            - ```sheet1.add_data_validation(ref="A1:C5", type="textLength", props=10)```
            - ```sheet1.add_data_validation(ref="A1:C5", type="list", props=["Yes", "No"])```
            - ```sheet1.add_data_validation(ref="A1:C5", type="whole", operator="between", props=[0, 100])```
        """
        if isinstance(ref, str):
            ref = [ref]
        ref = [self.util_range_ref_from_coordinates(self.util_get_range_coordinates(i, enrich_dimensions=True)) if i in self.header else i for i in ref]
        super().add_data_validation(ref=ref, type=type, props=props, operator=operator, **kwargs)


    def util_get_cell_coordinates(self, ref):
        """
        Get cell coordinates

        Examples:
            >>> ws.util_get_cell_coordinates(ref='employees')
            (N, 3)
            >>> ws.util_get_cell_coordinates(ref='A2')
            (2, 1)
            >>> ws.util_get_cell_coordinates(ref='A')
            (None, 1)
            >>> ws.util_get_cell_coordinates(ref='1')
            (1, None)
            >>> ws.util_get_cell_coordinates(ref='-1')
            (-1, None)
        """
        ref = str(ref)
        if not ref.isdigit() or not (len(ref) >= 2 and ref[0] == '-' and ref[1:].isdigit()):
            if ref in self.header:
                ref = openpyxl.utils.cell.get_column_letter((self.header.get_loc(ref) + 1) + self.index_depth + self.startcol)
        return super().util_get_cell_coordinates(ref)


class ExcelWriter:
    """
    Class for writing DataFrame objects into excel sheets.

    Example:
        Output pandas dataframe quickly with beautiful formatting.

        ```python
        from beautifulexcel import ExcelWriter

        with ExcelWriter('workbook.xlsx', mode='r', style='elegant_blue') as writer:
            ws1 = writer.to_excel(df1, sheetname='My Sheet', mode='a', startrow=0, startcol=0)
            ws2 = writer.to_excel(df2, sheetname='My Sheet', mode='a', startrow=20, startcol=0)
        ```
    """

    def __init__(
        self,
        file: str,
        mode: str = "replace", #Literal["replace", "modify"] = "replace",
        if_sheet_exists: str = "replace", #Literal["error", "new", "replace", "overlay"] = "replace",
        theme: str = "elegant_blue",
        ref_warnings: bool = True,
        date_format: str = None,
        datetime_format: str = None,
        engine_kwargs: Any = {},
        **kwargs,
    ):
        """
        Args:
            file (str): Path to xls or xlsx or ods file
            mode (str): If the file already exists you can either "replace" or "modify" it
            if_sheet_exists (str): If a excel sheet already exists raise an "error", create a "new" sheet with a different name, "replace" the existing sheet with the new one, or "overlay" the new contents with the old ones
            theme (str): Excel style name or path to theme yaml file
            date_format (str): Format string for dates written into Excel files (e. g. 'YYYY-MM-DD')
            datetime_format (str): Format string for datetime objects written into Excel files. (e. g. 'YYYY-MM-DD HH:MM:SS')
            engine_kwargs (str): keywords passed though to openpyxl in "replace"-mode: openpyxl.Workbook(**engine_kwargs); "modify"-mode: openpyxl.load_workbook(file, **engine_kwargs)

        Example:
            ```python
            from beautifulexcel import ExcelWriter

            with ExcelWriter('workbook.xlsx', mode='r', style='elegant_blue') as writer:
                ...
            ```
        """
        self.file = file
        self.mode = mode
        self.if_sheet_exists = if_sheet_exists
        self.theme_name = theme
        self.ref_warnings = ref_warnings
        self.date_format = date_format
        self.datetime_format = datetime_format
        self.engine_kwargs = engine_kwargs

        # modify existing file
        if os.path.isfile(file) and mode == "modify":
            self.writer = pd.ExcelWriter(
                file,
                engine="openpyxl",
                mode="a",
                if_sheet_exists=if_sheet_exists,
                date_format=date_format,
                datetime_format=datetime_format,
                engine_kwargs=engine_kwargs,
                **kwargs,
            )
            # self.writer.book = openpyxl.load_workbook(file, **engine_kwargs)
            # self.writer.sheets = {ws.title: ws for ws in self.writer.book.worksheets}
            self.file_mode = "modify"

        # create new file
        else:
            self.writer = pd.ExcelWriter(
                file,
                engine="openpyxl",
                mode="w",
                if_sheet_exists=None,
                date_format=date_format,
                datetime_format=datetime_format,
                engine_kwargs=engine_kwargs,
                **kwargs,
            )
            # self.writer.book = openpyxl.Workbook(**engine_kwargs)
            self.file_mode = "replace"

        # explicitly no theme defined
        if theme is None or len(theme) == 0:
            self.theme = {}

        # theme defined
        else:
            # standard theme from package
            if "." not in theme:
                this_file_path = os.path.dirname(os.path.abspath(__file__))
                theme = os.path.join(this_file_path, "themes", f"{theme}.yml")

            # read in user yml theme file
            try:
                with open(theme, "r") as file:
                    theme = yaml.safe_load(file)
                    theme_converted = {}
                    # flatten after level 2
                    for level1_name, level1 in theme.items():
                        if level1_name not in theme_converted:
                            theme_converted[level1_name] = {}
                        for level2_name, level2 in level1.items():
                            if level2_name not in theme_converted:
                                theme_converted[level1_name][level2_name] = {}
                            theme_converted[level1_name][level2_name] = flatten_dict(level2)
                    self.theme = theme_converted
            except yaml.YAMLError as exc:
                raise Exception(f"Error when reading in theme file from path '{theme}':", exc)

    def __enter__(self):
        if not hasattr(self, "file"):
            raise Exception(
                "Wrong usage! Please run it this way:\n>>> from beautifulexcel import ExcelWriter\n>>> with ExcelWriter('workbook.xlsx', mode='r', theme='elegant_blue') as writer:\n>>>     ws1 = writer.write_df(df1, sheetname='My Sheet'"
            )
        return self

    def save(self):
        if callable(getattr(self.writer, "save", None)):
            self.writer.save()
        else:
            self.writer._save()

    def __exit__(self, type, value, traceback):
        # ToDo: add exception handling here
        self.save()
        self.writer = None
        self = None

    def to_excel(
        self,
        df: pd.DataFrame,
        sheet_name: str,
        startrow: int = 0,
        startcol: int = 0,
        index: bool = False,
        header: bool = True,
        style: dict = {},
        use_base_style: bool = True,
        col_widths: dict = {},
        col_autofit: bool = True,
        auto_number_formatting: bool = True,
    ) -> DataframeSheet:
        """
        Export pandas Datafame to excel.

        Args:
            df (pd.DataFrame): Pandas Dataframe to export
            sheet_name (str): Sheet name
            startrow (int): Upper left cell row to dump dataframe (zero indexed)
            startcol (int): Upper left cell column to dump data rame (zero indexed)
            index (bool): Write row names/index as first column/s
            header (bool): Write column names/header as first row/s
            style (dict): Style dictionary with key referencing the cell/column/row and value the style to apply e.g. {'RoE': 'bg_light_blue', 'D:E': {'fill': 'FFEEB7'}}
            use_base_style (bool): Apply the excel workbook "theme" set in ExcelWriter()
            col_widths (dict): Define column widths manually with key referencing the column and value the width e.g. {'A:C': 20, 'F': 10, 'employees': 40}
            col_autofit (bool): Automatically change column width to fit content best
            auto_number_formatting (bool): Automatically detect number format and change excel format

        Returns:
            beautifulexcel.DataframeSheet

        Example:
            ```python
            from beautifulexcel import ExcelWriter

            with ExcelWriter('workbook.xlsx', mode='r', style='elegant_blue') as writer:
                ws1 = writer.to_excel(df1, sheetname='My Sheet', mode='a', startrow=0, startcol=0)
            ```
        """
        df_sheet = DataframeSheet(
            excelwriter=self,
            df=df,
            sheet_name=sheet_name,
            startrow=startrow,
            startcol=startcol,
            index=index,
            header=header,
            style=style,
            use_theme_style=use_base_style,
            col_widths=col_widths,
            col_autofit=col_autofit,
            auto_number_formatting=auto_number_formatting,
        )

        return df_sheet


if __name__ == "__main__":
    example_df = pd.DataFrame(
        {
            "client": ["A", "B", "C", "D", "E", "F", "G", "H"],
            "industry": [
                "ASEET MANAGEMENT",
                "ASEET MANAGEMENT",
                "BANK",
                "INSURANCE",
                "VERY VERY VERY VERY VERY LONG INSUSTRY NAME",
                "BANK",
                "COMMODITY BROKER",
                "INSURANCE",
            ],
            "employees": [25_000, 17_000_000, 14, 9_000, 12_000_000, 9_000, np.nan, 4_000_000],
            "inception": [
                datetime.datetime(2022, 1, 1),
                datetime.datetime(2000, 10, 30),
                datetime.datetime(2010, 5, 6),
                np.nan,
                datetime.datetime(1997, 1, 1),
                datetime.datetime(1962, 1, 1),
                datetime.datetime(2003, 11, 10),
                datetime.datetime(2022, 1, 1),
            ],
            "last_contact": [
                datetime.datetime(2022, 1, 1, 10, 2, 5),
                datetime.datetime(2000, 10, 30),
                datetime.datetime(2010, 5, 6),
                np.nan,
                datetime.datetime(1997, 1, 1),
                datetime.datetime(1962, 1, 1),
                datetime.datetime(2003, 11, 10, 10, 2, 5),
                datetime.datetime(2022, 1, 1),
            ],
            "RoE": [0.05, -0.05, 0.15, 1.05, -0.02, np.nan, 0.08, 0.05],
            "revenue": [
                50_000_000.4387,
                np.nan,
                63_000_000.4387,
                25_000.4387,
                25_000_000.4387,
                -50_000_000.4387,
                76_000_000.4387,
                25_000.4387,
            ],
        }
    )
    example_df.set_index("client", inplace=True)

    with ExcelWriter("workbook.xlsx", mode="r", theme="elegant_blue") as writer:
        ws1 = writer.to_excel(
            example_df,
            style={"RoE": "bg_light_blue", "D:E": {"fill": "FFEEB7"}},
            sheet_name="My Sheet",
            startrow=0,
            startcol=0,
            index=True,
            col_widths={"employees": 100},
        )
        ws1.add_data_validation(ref="employees", type="list", props=["Y", "N"])
        ws1.group_columns("revenue:B")
        ws1.merge_cells("B8:C9")
        ws1.write_cells({'A1':'Hello'})

    # example_df.to_excel('test.xlsx', sheet_name='My Sheet', startrow=0, startcol=0, index=True)
